"""Reworked PySimpleGUI's hello world with the PSGA feature"""

import PySimpleGUI as sg

import psga


# PSGA: define an action for the Ok button
@psga.action()
# PSGA: The action generates a unique key value (if the decorator has no explicit key argument)
def _on_ok(values):
    # PSGA: The action has a handler: simply print a value
    print("You entered ", values[0])


sg.theme("DarkAmber")  # Add a touch of color
# All the stuff inside your window.
layout = [
    [sg.Text("Some text on Row 1")],
    [sg.Text("Enter something on Row 2"), sg.InputText()],
    [
        sg.Button("Exit"),
        # PSGA: assign the action's key value to the element (with intellisense no typo mismatches)
        sg.Button("Ok", key=_on_ok.key),
    ],
]

# Create the Window
window = sg.Window("Window Title", layout)

# PSGA: initiate the dispatcher, register the action and loop (until the Exit event is read)
psga.Dispatcher().register(_on_ok).loop(window)

window.close()



from typing import Callable
from unittest.mock import MagicMock

import psga


def test_action_implicit_key():
    @psga.action()
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "key")
    assert getattr(my_action, "key") == my_action.key


def test_action_explicit_key():
    @psga.action(key="my_action")
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "key")
    assert getattr(my_action, "key") == "my_action"
    assert getattr(my_action, "key") == my_action.key


def test_dispatcher_register_one():
    handler1_invoked = 0

    @psga.action(key="event")
    def hander1(_):
        nonlocal handler1_invoked
        handler1_invoked += 1

    dispatcher = psga.Dispatcher()
    dispatcher.register(hander1)

    assert not dispatcher.dispatch("wrong_event", {})
    assert handler1_invoked == 0

    assert dispatcher.dispatch("event", {})
    assert handler1_invoked == 1

    assert dispatcher.dispatch(hander1.key, {})
    assert handler1_invoked == 2


def test_dispatcher_register_multiple():
    handler1_invoked = handler2_invoked = 0

    @psga.action()
    def handler1(_):
        nonlocal handler1_invoked
        handler1_invoked += 1

    @psga.action(handler1.key)
    def handler2(_):
        nonlocal handler2_invoked
        handler2_invoked += 1

    dispatcher = psga.Dispatcher()
    dispatcher.register(handler1).register(handler2)

    assert not dispatcher.dispatch("wrong_event", {})
    assert handler1_invoked == 0
    assert handler2_invoked == 0

    assert dispatcher.dispatch(handler1.key, {})
    assert handler1_invoked == 1
    assert handler2_invoked == 1

    assert dispatcher.dispatch(handler2.key, {})
    assert handler1_invoked == 2
    assert handler2_invoked == 2


def test_controller():
    class _MyController(psga.Controller):
        answer = 0

        @psga.action(key="universial_question")
        def on_ask(self, values):
            _MyController.answer = values

    dispatcher = psga.Dispatcher()
    controller = _MyController(dispatcher)

    assert controller.answer == 0
    dispatcher.dispatch(controller.on_ask.key, 42)
    assert controller.answer == 42


def test_dispatcher_loop():
    mock_window = MagicMock(return_value=[("Ok", {}), ("Exit", {})])

    mock_window = MagicMock()
    mock_window.configure_mock(
        **{
            "read.side_effect": [
                ("Ok", {}),
                ("Unknown", {}),
                (("-UNKNOWN TABLE KEY-", "+CICKED+", (3, 3)), {}),
                ("Unknown menu::unknown_menu_key", {}),
                ("Exit", {}),
            ],
        }
    )

    handler1_invoked = 0

    @psga.action(key="Ok")
    def hander1(_):
        nonlocal handler1_invoked
        handler1_invoked += 1

    dispatcher = psga.Dispatcher()
    dispatcher.register(hander1)

    dispatcher.loop(mock_window)

    assert handler1_invoked == 1



# Copyright 2023 Francis Meyvis <psga@mikmak.fun>

"""Minimalistic Controller (as in the MVC paradigm) for PySimpleGUI."""

import functools
from typing import Callable, Dict, Protocol, Self

import PySimpleGUI as sg


class Action(Protocol):
    """Bundles the event's name and handler"""

    key: str

    def __call__(self, values=None):
        """"""


def action(key: str | None = None):
    """Turns an event handler into an action using given key as event's name"""

    action._counter = getattr(action, "_counter", 0) + 1

    def _decorator_action(handler: Callable) -> Action:
        @functools.wraps(handler)
        def _wrapper_action(*args, **kwargs):
            handler(*args, **kwargs)

        _wrapper_action.key = handler.__name__ + "_" + str(action._counter) if key is None else key

        return _wrapper_action

    return _decorator_action


class Dispatcher:
    """Dispatcher an event's values to a matching handler."""

    def __init__(self):
        self._handlers: Dict[str, Action] = {}

    def register(self, handler: Action) -> Self:
        """Registers given action's handler by its key."""
        self._handlers.setdefault(handler.key, []).append(handler)
        return self

    def dispatch(self, event, values) -> bool:
        """Returns True if a handler was found and invoked for given event."""
        if isinstance(event, tuple):
            # TODO are there other type of "tuple"-events?
            key = event[0] if "+CLICKED+" == event[1] else None
        else:
            if 2 == len(menu_event := event.rsplit(sg.MENU_KEY_SEPARATOR, 1)):
                _, key = menu_event  # extract the key from a menu-item event having a key
            else:
                key = event

        if (handlers := self._handlers.get(key, None)) is not None:
            for handler in handlers:
                handler(values)
            return True
        return False

    def loop(self, window: sg.Window):
        """Process window's events and values until the Exit event."""
        while True:
            event, values = window.read()
            print(f"#### LOOP event: {event} values: {values}\n")

            if event in {sg.WIN_CLOSED, "Exit"}:
                break

            if self.dispatch(event, values):
                continue

            print(f"Unhandled event: {event}, values: {values}")


class Controller:
    """Groups and registers actions to a dispatcher."""

    def __init__(self, dispatcher: Dispatcher, window: sg.Window | None = None):
        self.window = window
        for method in [
            func
            for func in map(lambda name: getattr(self, name), dir(self))
            if callable(func) and hasattr(func, "key")
        ]:
            dispatcher.register(method)
